/*
 * Copyright (C) 2004, 2008 The Abersim group
 * Copyright (C) 2004, 2008 Department of Circulation and Medical Imaging
 * Copyright (C) 2004, 2008 Norwegian University of Science and Technology
 *
 * This file is a part of the Abersim computer package
 *
 * This Abersim computer package is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This Abersim computer package is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, please download it from
 * <http://www.gnu.org/licenses>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "abersim2.h"
//#include "./correction/abersim_correction.h"
//TODO make this extern thing work properly (place it in abersim_correction.h somehow)
//extern int test(int argc, char **argv);

void printUsage() {
	printf("\n----------------------- Abersim 2.0 Usage -----------------------\n\n");
	printf("For aberration simulation include filename containing MATLAB data.\n");
	printf("\t./abersim2 mfile\n\n");
	printf("For aberration correction include filename containing MATLA data, \n");
	printf("as well as the center frequency index in the FFT, the bandwidth \n"); 
	printf("around the FC in the FFT, the X-, Y- and Z-dimensions. Optionally \n");
	printf("include number of streams to use, the number of iterations for \n");
	printf("the estimation loop and the device to use.\n");
	printf("\t./abersim2 mfile 25 14 100 100 512 3 1 1\n\n");
}

int main (int argc, char **argv) {
	int nstr;
	char *instr, *fnstr, *logstr;

	// Find file name
	if ( argc > 1 ) {
		instr = argv[1];
	} else {
		printUsage();
		return 0;
	}
	nstr = strlen(instr);
	logstr = malloc( (nstr+5) * sizeof(char) );
	fnstr =  malloc( (nstr+5) * sizeof(char) );
	sprintf(logstr, "%s.log", instr);
	sprintf(fnstr,  "%s.mat", instr);

	char ans;
	printf("press S for simulation and C for correction: ");
	scanf("%c", &ans);
	if (ans == 's' || ans == 'S') 
	{
		// Run simulation
		init_log(logstr, instr, 1);
		start_log(logstr);
		
		plan_and_run_simulation(fnstr, logstr);
		
		close_log(logstr);
	}
	else
	{
		//Get FC, BANDWIDTH, XSIZE, YSIZE, ZSIZE and n from arguments
		int fc_index,bandwidth_indexes,xsize,ysize,zsize,streams,n,device;
		if ( argc > 6 ) {
		if(!(fc_index=atoi(argv[2]))){ return 1; }
//		if(!(bandwidth_indexes=atoi(argv[3]))){ return 1; }
		bandwidth_indexes=atoi(argv[3]); //Illegal values -> 0
		if(!(xsize=atoi(argv[4]))){ return 1; }
		if(!(ysize=atoi(argv[5]))){ return 1; }
		if(!(zsize=atoi(argv[6]))){ return 1; }
		if (argc > 9) {	if(!(device=atoi(argv[9]))){ device = 0; } } 
		if (argc > 8) {	if(!(n=atoi(argv[8]))){ n = 1; } } 
		if (argc > 7) {	if(!(streams=atoi(argv[7]))){ streams = 3; } }
		} else {
			printf("ERROR: To few arguments.\n");
			printUsage();
			return 1;
		}
		
		// Run correction
		init_log(logstr, instr, 1);
		start_log(logstr);
		
		printf("\n");
		fftw_complex *indata;
		cufftComplex *result;
		int harmonic = 1;
		
		//Set device to use.
		initializeCuda(device);
		//Allocate indata and result matrices.
		allocate(&indata, &result, xsize, ysize, zsize, streams);
		//Load indata from file.
		loadData(indata, fnstr, logstr, xsize, ysize, zsize, "u_z");
		
		//Estimate and correct aberration.
		abcorrection(indata, result, fc_index, bandwidth_indexes, xsize, ysize, zsize, streams, n);
		
		//Save result to file.
		saveData(result, fnstr, logstr, xsize, ysize, zsize, harmonic, "amplitude", "phase");
		//De-allocate matrices used.
		deAllocate(indata, result);
				
		close_log(logstr);
	}
	free(fnstr);
	free(logstr);
	return 0;
}

