#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

//#include <cutil.h>

extern "C" {
#include "../general/abersim_general.h"
#include "../general/propcontrol.h"
#include "../math/abersim_math.h"
#include "../io/abersim_io.h"
//#include "../debug/debug.h"
//#include "../aberration/abersim_aberration.h"
#include "abersim_correction.h"
//#include "cpuCast.h"
}

#include "convolutionS.h"
#include "gpuOperations.h"
#include "cpuCorrelation.h"
//#include "gpuCorrelation.h"
#include "cpuFFT.h"
#include "gpuFFT.h"
//#include "convolutionSeparable.h"
#include "gpuCast.h"


////////////////////////////////////////////////////////////////////////////////
// CONFIGURATION
////////////////////////////////////////////////////////////////////////////////
const int N = 50; 						// Number of times to loop convolution kernels - higher better convergence?
const float MU = 0.1f;					// Convergence factor
const float EPS = 0.01f;				// Error limit, epsilon
const float T = 1; //101				// Number of loops to run to collect test data

// Code that gives very accurate timer.
static unsigned long long rdtsctime()
{
    unsigned int eax, edx;
    unsigned long long val;
    __asm__ __volatile__("rdtsc":"=a"(eax), "=d"(edx));
    val = edx;
    val = val << 32;
    val += eax;
    return val;
}

extern "C" {
void abcorrection(	fftw_complex *src, 
					cufftComplex *rM_h, 
					const int fc, 
					const int bandwidth, 
					const int xsize, 
					const int ysize, 
					const int zsize, 
					const int streams, 
					const int n) {
	
	printf(" -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- \n");
	printf(" -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- \n");
	printf(" -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- NEW RUN -- \n");
	
	
	////////////////////////////////////////////////////////////////////////////////
	// VARIABLES
	////////////////////////////////////////////////////////////////////////////////
	int i, j, k, t=0;
	//Error holder
	cudaError err;
	//Timer variables
	unsigned long long timestart, timefinish, totstart, totfinish, loopstart, loopfinish;
	//Number of blocks to execute for kernel and number of threads in each block.
	dim3 nBlocks, blockSize;
	//Stores number of bytes used for memcpy
	size_t bytes;
	int totsize;
	totsize = xsize*ysize*zsize;
	
	
	
	////////////////////////////////////////////////////////////////////////////////
	//ALLOCATIONS (These are only done once per program run)
	////////////////////////////////////////////////////////////////////////////////
	timestart = rdtsctime();

	fftwf_complex *fsrc; //float version of fftw complex numbers.
	fftw_complex *cudaSrc;
	cudaMallocHost((void **) &fsrc, streams*totsize * sizeof(fftwf_complex));
	cudaMalloc((void **) &cudaSrc, streams*totsize * sizeof(fftw_complex));
	
	//allocation of matrices used for FFTW.
	fftwf_complex ***cpueM, ***cpurM; //elementMatrix and resultMatrix
	cpueM = (fftwf_complex***) malloc(sizeof(fftwf_complex**)*xsize);
	for (i = 0; i < xsize; i++) {
		cpueM[i] = (fftwf_complex**) malloc(sizeof(fftwf_complex*)*ysize); 
		for (j = 0; j < ysize; j++) {
			cpueM[i][j] = (fftwf_complex*) malloc(sizeof(fftwf_complex)*zsize);
		}
	}
	cpurM = (fftwf_complex***) malloc(sizeof(fftwf_complex**)*xsize);
	for (i = 0; i < xsize; i++) {
		cpurM[i] = (fftwf_complex**) malloc(sizeof(fftwf_complex*)*ysize); 
		for (j = 0; j < ysize; j++) {
			cpurM[i][j] = (fftwf_complex*) malloc(sizeof(fftwf_complex)*zsize);
		}
	}
	fillRandom(cpueM, xsize,ysize,zsize);
	fillRandom(cpurM, xsize,ysize,zsize);
	setupFFT(zsize, xsize, ysize, zsize, cpueM[0][0], cpurM[0][0]);
	
	//Allocation of matrices used for CPU kernel.
	cufftComplex **elementMatrix_h, **correlationMatrix_h;
	createMatrix(&elementMatrix_h,xsize,ysize);
	createMatrix(&correlationMatrix_h,xsize,ysize);	
	
	
	
	//GPU matrices
	cufftComplex *elementMatrix, *cufftResultMatrix, *tempMatrix_h;
	cudaMalloc((void**) &elementMatrix, sizeof(cufftComplex)*streams*totsize); //Data to CUFFT
	cudaMalloc((void**) &cufftResultMatrix, sizeof(cufftComplex)*streams*totsize); //Result from CUFFT
	cudaMallocHost((void**) &tempMatrix_h, sizeof(cufftComplex)*streams*totsize); //PRINTF matrix //TODO remove 
	cuda_setupFFT(zsize, xsize*ysize);
	
	cufftComplex *s_h, *sOLD_h;	
	cudaMallocHost((void**) &s_h, sizeof(cufftComplex)*streams*xsize*ysize);
	cudaMallocHost((void**) &sOLD_h, sizeof(cufftComplex)*streams*xsize*ysize);

	cufftComplex *eM_d, *sM_d, *sOM_d, *rM_d;
	//sPitch gives size of padded array in bytes.
	//sPitch/8 gives number of elements in padded array (of cufftComplex of doubles).
	size_t ePitch;
	size_t sPitch;
	size_t sOPitch;
	size_t rPitch;
	//cudaMallocHost allocates page-locked memory on host (enables cudaMemcpy2DAsync and other functionality).
	//Allocate strided memory that is aligned, speeds up memory transfers and access.
	cudaMallocPitch((void **) &eM_d, &ePitch, xsize * sizeof(cufftComplex), streams*ysize); //Data for convolution
	cudaMallocPitch((void **) &sM_d, &sPitch, xsize * sizeof(cufftComplex), streams*ysize); //Correctionfactors that change ever iteration of convolution
	cudaMallocPitch((void **) &sOM_d, &sOPitch, xsize * sizeof(cufftComplex), streams*ysize); //Original correctionfactors
	cudaMallocPitch((void **) &rM_d, &rPitch, xsize * sizeof(cufftComplex), streams*ysize); //Result from convolution
	//TODO remember to use pitch when addressing eM_d, sM_d, sOM_d and rM_d!!!! ePitch = sPitch = sOPitch = rPitch = 64 (/8 = 8)
	
	//Create streams to interleave processing
	cudaStream_t stream[streams];
	for (i = 0; i < streams; i++) {
		cudaStreamCreate(&stream[i]);
	}
	
	timefinish = rdtsctime();
	printf("\nALLOCATE: Time to allocate (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));

	
	////////////////////////////////////////////////////////////////////////////////
	// LOOP FOR EACH STREAM
	////////////////////////////////////////////////////////////////////////////////	
	loopstart = rdtsctime();
	for (t = 0; t < T; t++) { //TODO remove
//		printf("%d\t",t);
		printf("////////////////////////////////////////////////////////////////////////////////\n");
		printf("ROUND %d\n", t);
		printf("////////////////////////////////////////////////////////////////////////////////\n");
	
	//resetting variables
	i = j = k = 0;
	int blocks, threads;
	fillRandom(elementMatrix_h, xsize, ysize);
	//Set all matrices to start values //TODO remove
//	int blocks = iDivUp(totsize, 512*30) * 30;
//	if (blocks > 512) blocks = 512;
//	int threads = 512;
//	if (totsize < 512) threads = totsize; 
//	nBlocks = dim3(blocks);
//	blockSize = dim3(threads);
//	cuda_fill<<<nBlocks, blockSize>>>(elementMatrix, totsize, 0.0f); //TODO remove
//	cuda_fill<<<nBlocks, blockSize>>>(cufftResultMatrix, totsize, 0.0f); //TODO remove

	nBlocks = dim3(streams*ysize,xsize); //TODO remove
//	cuda_clearMatrix<<<nBlocks,1>>>(eM_d, ePitch); //TODO remove
//	cuda_clearMatrix<<<nBlocks,1>>>(rM_d, rPitch); //TODO remove
	cuda_setStartValues<<<nBlocks,1>>>(sM_d, sPitch);
//	cuda_setStartValues<<<nBlocks,1>>>(sOM_d, sOPitch); //TODO remove	
	
	//Stream to run current loop on.
	int s = t % streams;
		
	
	
	totstart = rdtsctime();
	////////////////////////////////////////////////////////////////////////////////
	// COPY DATA TO DEVICE
	////////////////////////////////////////////////////////////////////////////////
	printf("\n");
	timestart = rdtsctime();
	//Assuming non-page-locked memory from connected code so we cannot use Async mode
	bytes = sizeof(fftw_complex)*totsize;
	cudaMemcpy(	(char*)cudaSrc + s*totsize*sizeof(fftw_complex), 
				(char*)src, 
				bytes, 
				cudaMemcpyHostToDevice);
	cudaThreadSynchronize(); //Need to synchronize to get correct time count.
	timefinish = rdtsctime();
	printf("COPY: Time to copy data to device (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	err = cudaGetLastError();
	if (err != cudaSuccess)	printf("ERROR: %s :ERROR\n", cudaGetErrorString( err ) );
	
	
	
	////////////////////////////////////////////////////////////////////////////////
	// CASTING USING CUDA
	////////////////////////////////////////////////////////////////////////////////
	//Casting from a complex double to a complex float.
	timestart = rdtsctime();
	printf("Casting using CUDA...\n");
	blocks = iDivUp(totsize*2, 512*30) * 30;
	if (blocks > 512) blocks = 512;
	threads = 512;
	if (totsize*2 < 512) threads = totsize*2; 
		
	nBlocks = dim3(blocks);
	blockSize = dim3(threads);
	
	double* dcudaSrc = (double*) cudaSrc;
	float* feM = (float*) elementMatrix;
	cuda_castFromDouble<<<nBlocks, blockSize, 0, stream[s]>>>(	(double*)((char*)dcudaSrc + s*sizeof(double)*totsize), 
																(float*)((char*)feM + s*sizeof(float)*totsize), 
																totsize*2);
	cudaThreadSynchronize(); //Need to synchronize to get correct time count.
	timefinish = rdtsctime();
	printf("CAST: Time to cast using CUDA (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	err = cudaGetLastError();
	if (err != cudaSuccess)	printf("ERROR: %s :ERROR\n", cudaGetErrorString( err ) );
	
	
	
	////////////////////////////////////////////////////////////////////////////////
	// Compute FFT using CUFFT
	////////////////////////////////////////////////////////////////////////////////
	printf("\n");
	timestart = rdtsctime();
	printf("Computing FFT using CUFFT...\n");
	cuda_doFFT((cufftComplex*)((char*)elementMatrix + s*sizeof(cufftComplex)*totsize), (cufftComplex*)((char*)cufftResultMatrix + s*sizeof(cufftComplex)*totsize));
	cudaThreadSynchronize(); //Need to synchronize to get correct time count.
	timefinish = rdtsctime();
	printf("CUFFT: Time for CUFFT to compute (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	err = cudaGetLastError();
	if (err != cudaSuccess)	printf("ERROR: %s :ERROR\n", cudaGetErrorString( err ) );
	
	
	
	////////////////////////////////////////////////////////////////////////////////
	//GPU CONVOLUTION
	////////////////////////////////////////////////////////////////////////////////
	printf("\n");
	timestart = rdtsctime();
	
/*	float *minError_d, *minError_h;
	cudaMallocHost((void**) &minError_h, sizeof(float));
	cudaMalloc((void**) &minError_d, sizeof(float));
	minError_h[0] = 100000000000.0f;
	cudaMemcpy(	minError_d, minError_h, sizeof(float), cudaMemcpyHostToDevice);
*/	
	
	float minError;
	minError = 100000000000.0f;

	
	cudaThreadSynchronize();
	cudaMemcpy2DAsync(	s_h, xsize*sizeof(cufftComplex), 
						sM_d, sPitch, 
						xsize*sizeof(cufftComplex), ysize, 
						cudaMemcpyDeviceToHost, stream[s]);
	cudaThreadSynchronize();
	
	threads = floor(sqrt(16384 / ((bandwidth+1) * 8 + 8)));
	if (threads-KERNEL_W < 1) {
		printf("FATAL ERROR: Bandwidth to wide.\n");
		exit(1);
	}
	else if (threads*threads*50 > 16199) {
		threads = 17; //Uses maximum registers available per SM.
	}
	if ((threads*threads*(bandwidth+1)+threads*threads)*8 > 16383) {
		threads = 15; //Uses maximum shared memory available per block.		
	}
//	printf("threads: %d \n",threads);
//	printf("resources: %d\n", (threads*threads*(bandwidth+1)+threads*threads)*8);
	
	nBlocks = dim3(iDivUp(xsize+KERNEL_W,threads-KERNEL_W), iDivUp(ysize+KERNEL_W,threads-KERNEL_W)); //7 is maximum of threads 11 - KERNEL_RADIUS*2
	blockSize = dim3(threads,threads); // (11,11)
	
	printf("Computing new correction matrix...\n");
	i = 0;
	do {
		convolutionS<<<nBlocks, blockSize, (threads*threads*(bandwidth+1)+threads*threads)*8, stream[s]>>>(	
															(cufftComplex*)((char*)cufftResultMatrix + s*sizeof(cufftComplex)*totsize), 
															(cufftComplex*)((char*)rM_d + s*rPitch*ysize), 
	 														(cufftComplex*)((char*)sM_d + s*sPitch*ysize),
	 														MU, 
	 														xsize, 
	 														ysize, 
	 														zsize,  
															sPitch, 
															bandwidth, 
															fc);
		
		//Copy current S to old S for error calculation
		for (j = 0; j < ysize; j++) {
			for (k = 0; k < xsize; k++) {
				sOLD_h[k+j*xsize].x = s_h[k+j*xsize].x;
				sOLD_h[k+j*xsize].y = s_h[k+j*xsize].y;
			}
		}
		
		cudaThreadSynchronize();
		//Copy new S to current S for error calculation
		cudaMemcpy2DAsync(	s_h, xsize*sizeof(cufftComplex), 
							rM_d, rPitch, 
							xsize*sizeof(cufftComplex), ysize, 
							cudaMemcpyDeviceToHost, stream[s]);
		cudaThreadSynchronize();
		
		//Calculate Error using CPU
		float temp = 0;
		for (j = 0; j < ysize; j++) {
			for (k = 0; k < xsize; k++) {
				temp += sqrt(sOLD_h[k+j*xsize].x*sOLD_h[k+j*xsize].x + sOLD_h[k+j*xsize].y*sOLD_h[k+j*xsize].y) / sqrt(s_h[k+j*xsize].x*s_h[k+j*xsize].x + s_h[k+j*xsize].y*s_h[k+j*xsize].y);
			}
		}
		temp /= ysize*xsize;
		temp = abs(temp - 1.0f);
		if (temp < minError) {
			minError = temp;
		}
//		printf("minError: %f\n",minError);
//		printf("Iteration: %d\n",i);
		
		cudaThreadSynchronize();
		//Copy new S to current S for next iteration.
		cudaMemcpy2DAsync(	(cufftComplex*)((char*)sM_d + s*sPitch*ysize), sPitch, 
							(cufftComplex*)((char*)rM_d + s*rPitch*ysize), rPitch, 
							xsize*sizeof(cufftComplex), ysize, 
							cudaMemcpyDeviceToDevice, stream[s]);
		cudaThreadSynchronize();
		
		i++;
//	} while (i < N && minError_h[0] > 0.01f);
	} while (i < N && minError > EPS);
	printf("Iterations used: %d\n",i+1);
	
	
	timefinish = rdtsctime();
	printf("GPU: Time for GPU to compute convolution (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	err = cudaGetLastError();
	if (err != cudaSuccess)	printf("ERROR: %s :ERROR\n", cudaGetErrorString( err ) );	
	
	
	
	printf("\n");
	////////////////////////////////////////////////////////////////////////////////
	// COPY RESULT BACK TO HOST
	////////////////////////////////////////////////////////////////////////////////
	timestart = rdtsctime();
	cudaMemcpy2DAsync(	rM_h, xsize*sizeof(cufftComplex), 
						sM_d, sPitch, 
						xsize*sizeof(cufftComplex), ysize, 
						cudaMemcpyDeviceToHost, stream[s]);
	timefinish = rdtsctime();
	printf("COPY: Time to copy result to host (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	err = cudaGetLastError();
	if (err != cudaSuccess)	printf("ERROR: %s :ERROR\n", cudaGetErrorString( err ) );
	
	
	
	totfinish = rdtsctime();
	printf("\nTOTAL: total time for GPU operations without allocations (on hpc6): %fms\n", 1000*((double)(totfinish-totstart)/2830000000));
//	printf("%f\t",1000*((double)(totfinish-totstart)/2830000000));
	
	

	
	

	
	
	totstart = rdtsctime();
	////////////////////////////////////////////////////////////////////////////////
	//Compute FFT using FFTW
	////////////////////////////////////////////////////////////////////////////////
	printf("\n");
	timestart = rdtsctime();
	printf("Copying in-data to fftw matrix strutured array (for FFTW)...\n");
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			for (k = 0; k < zsize; k++) {
				cpueM[i][j][k][0] = src[k+j*zsize+i*zsize*ysize][0];
				cpueM[i][j][k][1] = src[k+j*zsize+i*zsize*ysize][1];
			}
		}
	}
	timefinish = rdtsctime();
	printf("FFTW: Time to copy in-data to fftw matrix strutured array (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	timestart = rdtsctime();
	printf("Performing FFT using FFTW...\n");
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			doFFT(cpueM[i][j], cpurM[i][j]);
		}
	}
	timefinish = rdtsctime();
	printf("FFTW: Time for FFTW to compute (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));
	
	
	
	////////////////////////////////////////////////////////////////////////////////
	//CPU "CONVOLUTION"
	////////////////////////////////////////////////////////////////////////////////
	timestart = rdtsctime();
	printf("Computing correlation using CPU...\n");

	minError = 100000000000.0f;
	cufftComplex **errorMatrix_h;
	createMatrix(&errorMatrix_h,xsize,ysize);
	for (j = 0; j < xsize; j++) {
		for (k = 0; k < ysize; k++) {
			errorMatrix_h[j][k].x = elementMatrix_h[j][k].x;
			errorMatrix_h[j][k].y = elementMatrix_h[j][k].y;
		}
	}
	
	i = 0;
	do {
		correlateReduce(	cpurM, 
							elementMatrix_h, 
							correlationMatrix_h, 
							xsize, ysize, KERNEL_RADIUS, KERNEL_WF, MU, bandwidth, fc);
		
		//Copy current S to old S for error calculation.
		for (j = 0; j < xsize; j++) {
			for (k = 0; k < ysize; k++) {
				errorMatrix_h[j][k].x = elementMatrix_h[j][k].x;
				errorMatrix_h[j][k].y = elementMatrix_h[j][k].y;
			}
		}
		
		//Copy new S to current S for next iteration.
		for (j = 0; j < xsize; j++) {
			for (k = 0; k < ysize; k++) {
				elementMatrix_h[j][k].x = correlationMatrix_h[j][k].x;
				elementMatrix_h[j][k].y = correlationMatrix_h[j][k].y;
			}
		}
		
		//Error calculation
		float temp = 0;
		for (j = 0; j < xsize; j++) {
			for (k = 0; k < ysize; k++) {
				temp += sqrt(errorMatrix_h[j][k].x*errorMatrix_h[j][k].x + errorMatrix_h[j][k].y*errorMatrix_h[j][k].y) / sqrt(elementMatrix_h[j][k].x*elementMatrix_h[j][k].x + elementMatrix_h[j][k].y*elementMatrix_h[j][k].y);
			}
		}
		temp /= ysize*xsize;
		temp = abs(temp - 1.0f);
		if (temp < minError) {
			minError = temp;
		}

		i++;
	} while (i < N && minError > EPS);
	printf("Iterations used: %d\n",i+1);
	//Write to rM_h which is saved to file
	/*for (j = 0; j < XSIZE; j++) {
		for (k = 0; k < YSIZE; k++) {
			rM_h[k + j*XSIZE].x = correlationMatrix_h[k][j].x;
			rM_h[k + j*XSIZE].y = correlationMatrix_h[k][j].y;
		}
	}*/
	
	timefinish = rdtsctime();
	printf("CPU1: Time for CPU to compute (in seconds) (on hpc6): %f\n", ((double)(timefinish-timestart)/2830000000));

	totfinish = rdtsctime();
	printf("\nTOTAL: total time for CPU computations (on hpc6): %fms\n", 1000*((double)(totfinish-totstart)/2830000000));
//	printf("%f\n", 1000*((double)(totfinish-totstart)/2830000000));

	} //end for testing for-loop //TODO remove
	loopfinish = rdtsctime();
	printf("\nTOTAL: total time for loop including cast to float (on hpc6): %fms\n", 1000*((double)(loopfinish-loopstart)/2830000000));
	printf("\n");

	////////////////////////////////////////////////////////////////////////////////
	//DEALLOCATION (Only done once per program run)
	////////////////////////////////////////////////////////////////////////////////

	for(i = 0; i < xsize; i++) {
		for(j = 0; j < ysize; j++) {
			free(cpueM[i][j]);
		}
		free(cpueM[i]);
	}
	free(cpueM);
	for(i = 0; i < xsize; i++) {
		for(j = 0; j < ysize; j++) {
			free(cpurM[i][j]);
		}
		free(cpurM[i]);
	}
	//Free arrays used for error calculation.
	cudaFreeHost(s_h);
	cudaFreeHost(sOLD_h);	
	free(cpurM);
	destroyMatrix(&elementMatrix_h,ysize);
	destroyMatrix(&correlationMatrix_h,ysize);
	cudaFree(elementMatrix);
	cudaFree(cufftResultMatrix);
	cudaFree(eM_d);
	cudaFree(sM_d);
	cudaFree(sOM_d);
	cudaFree(rM_d);
	cuda_cleanupFFT();
	cleanupFFT();
	for (i = 0; i < streams; i++) {
		cudaStreamDestroy(stream[i]);
	}
}
} //extern "C"


























////////////////////////////////////////////////////////////////////////////////
// READ DATA FROM FILE
////////////////////////////////////////////////////////////////////////////////
void loadData(	fftw_complex *src, 
				char *fnstr, 
				char *logstr, 
				const int xsize, 
				const int ysize, 
				const int zsize, 
				char *matrixName) {
	
	double *u;
	long long int ret;
	int n1,n2,n3,x,y,z;
	n1 = xsize;
	n2 = ysize;
	n3 = zsize;
	
	u = (double*)malloc( n1*n2*n3 * sizeof(double) );
	for (y = 0; y < n1*n2*n3; y++) {
		u[y] = 0.0; //TODO remove (?)
	}
	
	//Read indata matrix from file.
	ret = read_genfile(n1, n2, n3, fnstr, matrixName, u);
	if ( ret ) {
		write_log(logstr, "Matrix may not be loaded fully");
	}
	
	//Change from Row major of MATLAB to column major.
	for (x = 0; x < n1; x++) {
		for (y = 0; y < n2; y++) {
			for (z = 0; z < n3; z++) {
				src[z+y*n3+x*n2*n3][0] = u[z*n1*n2+y+x*n1];
			}
		}
	}
	free(u);	
	write_log(logstr, "Signal loaded");
}

////////////////////////////////////////////////////////////////////////////////
// WRITE RESULT TO FILE
////////////////////////////////////////////////////////////////////////////////
void saveData(	cufftComplex *result, 
				char *fnstr, 
				char *logstr, 
				const int xsize, 
				const int ysize, 
				const int zsize, 
				int harmonic, 
				char *realName, 
				char *imagName) {
	
	int nh, n1, n2, x, y;
	nh = harmonic;
	n1 = xsize;
	n2 = ysize;
	
	long long int ret;
	double *real, *imag;
	real = (double*) malloc(n1*n2*sizeof(double));
	imag = (double*) malloc(n1*n2*sizeof(double));
	
	//Extract data from cufftComplex matrix into two seperate matrices
	for (y = 0; y < n2; y++) {
		for (x = 0; x < n1; x++) {
			real[x*n1+y] = result[x+y*n1].x;
			imag[x*n1+y] = result[x+y*n1].y;
		}
	}
	
	//Write real data to file
	ret = write_genfile(nh, n1, n2, fnstr, realName, real);
	if ( ret ) {
		write_log(logstr, "Real matrix saving failed.");
	}
	//Write imaginary data to file
	ret = write_genfile(nh, n1, n2, fnstr, imagName, imag);
	if ( ret ) {
		write_log(logstr, "Imaginary matrix saving failed.");
	}
	
	free(real);
	free(imag);
	
	write_log(logstr, "Matrices stored");
}


////////////////////////////////////////////////////////////////////////////////
// CUDA DEVICE INFORMATION
////////////////////////////////////////////////////////////////////////////////
void initializeCuda(int device) {
	//Select CUDA device to use.
	cudaDeviceProp deviceProp;
	cudaGetDeviceProperties(&deviceProp, device);
	printf("Using Device %d: %s\n", device, deviceProp.name);
	cudaSetDevice(device);
}

////////////////////////////////////////////////////////////////////////////////
// ALLOCATE MATRICES
////////////////////////////////////////////////////////////////////////////////
void allocate(	fftw_complex **src, 
				cufftComplex **rM_h, 
				const int xsize, 
				const int ysize, 
				const int zsize, 
				const int streams) {
	
	int totsize = xsize * ysize * zsize;
	cudaMallocHost((void **) src, totsize * streams * sizeof(fftw_complex)); //Allocate page-locked memory for faster access.
	fillRandom(src[0], totsize*streams); //TODO TODO TODO check why removing this will break program?
	
	cudaMallocHost((void **) rM_h, xsize * ysize * streams * sizeof(cufftComplex)); //Result from convolution copied to host //XSIZE*YSIZE
}

////////////////////////////////////////////////////////////////////////////////
// DEALLOCATE MATRICES
////////////////////////////////////////////////////////////////////////////////
void deAllocate(	fftw_complex *src, 
					cufftComplex *rM_h) {
	
	cudaFreeHost(src);
	cudaFreeHost(rM_h);
}





//TODO
//Implement everything as 12-bit (16-bit?) integers? including fft?
//using 3 streams to speed up - whole gpu acts as 3 seperate gpus and each run a  complete copy, calculate, copy process -pipeline like
//adapt to running with continues indata
//divide into multiple functions
