#include "gpuOperations.h"

////////////////////////////////////////////////////////////////////////////////
// Host functions
////////////////////////////////////////////////////////////////////////////////
/*
 * "Constructor" for matrices.
 */
void createMatrix(cufftComplex **matrix, int xsize, int ysize, size_t &pitch) {
	if (xsize == ysize) {
		//cudaMallocPitch inserts correct padding for CUDA
		cudaMallocPitch((void **) matrix, &pitch, xsize * sizeof(cufftComplex), ysize);
	}
	else {
		//Allocate a 1D array
		cudaMalloc((void **) matrix, sizeof(cufftComplex) * xsize * ysize);
	}
}

/*
 * "Destructor" for matrices.
 */
void destroyMatrix(cufftComplex **matrix) {
	cudaFree(matrix);
}

/*
 * "Constructor" for matrices.
 */
void createMatrix(cufftComplex ***matrix, int xsize, int ysize) {
	int i;
	*matrix = (cufftComplex**) malloc(sizeof(cufftComplex*) * xsize);
	for (i = 0; i < ysize; i++) {
		(*matrix)[i] = (cufftComplex*) malloc(sizeof(cufftComplex) * ysize);
	}
}

/*
 * "Destructor" for used matrices.
 */
void destroyMatrix(cufftComplex ***matrix, int ysize) {
	int i;
	for (i = 0; i < ysize; i++) {
		free((*matrix)[i]);
	}
	free(*matrix);
}

/*
 * Prints the matrix to screen
 */
void printMatrix(cufftComplex *matrix, int xsize, int ysize) {
	int i, j;//,t=0;
	/*for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
//			printf("(%.2f,%.0f)\t", matrix[t].x, matrix[t].y);
//			t++;
			printf("(%.2f,%.0f)\t", matrix[j+ysize*i].x, matrix[j+ysize*i].y);
			//printf("(%f,%f) ", matrix[i+ysize*j].x, matrix[i+ysize*j].y);
		}
		printf("\n");
	}*/
	/*for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			//printf("(%.2f,%.0f)\t", matrix[i+pitch*j].x, matrix[i+pitch*j].y);
			printf("(%.10f,%.10f)\t", matrix[j+ysize*i].x, matrix[j+ysize*i].y);
			//T* pElement = (T*)((char*)BaseAddress + Row * pitch) + Column;
		}
		printf("\n\n");
	}*/
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
			//printf("(%.2f,%.0f)\t", matrix[i+pitch*j].x, matrix[i+pitch*j].y);
			printf("(%.10f,%.10f)\t", matrix[j+xsize*i].x, matrix[j+xsize*i].y);
			//T* pElement = (T*)((char*)BaseAddress + Row * pitch) + Column;
		}
		printf("\n\n");
	}
}

/*
 * Prints the matrix to screen
 */
void printMatrix(cufftComplex *matrix, int xsize, int ysize, int zsize) {
	int i, j, k,t =0;
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
//			printf("(%d,%d): ",j,i);
			for (k = 0; k < zsize; k++) {
//				//printf("%d:", k+ysize*j+xsize*i);
//				printf("%d:", t);
				printf("(%f,%f)\t", matrix[t].x, matrix[t].y);
//				printf("%d, %f\n", t, matrix[t].x);
				t++;
			}
			printf("\n");
		}
		printf("\n");
	}
}

void printMatrix(	cufftComplex *matrix, int size) {
	int i;
	for (i = 0; i < size; i++) {
		printf("(%f,%f)\t", matrix[i].x, matrix[i].y);
	}
}
/*
 * Prints the matrix to screen
 */
void printMatrix(cufftComplex **matrix, int xsize, int ysize) {
	int i, j;
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			printf("(%f,%f) ", matrix[i][j].x, matrix[i][j].y);
		}
		printf("\n");
	}
}

/*
 * Prints the matrix to screen
 */
void printMatrix(fftwf_complex ***matrix, int xsize, int ysize, int zsize) {
	int i, j, k;
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
			printf("(%d,%d):",j,i);
			for (k = 0; k < zsize; k++) {
//				printf("%d,%d,%d:\n",i,j,k);
				printf("(%f,%f)\t", matrix[j][i][k][0], matrix[j][i][k][1]);
			}
			printf("\n");
		}
		printf("\n");
	}
}

/*
 * Prints the matrix to screen
 */
void printMatrix(fftw_complex *matrix, int xsize, int ysize, int zsize) {
	int i, j, k,t=0;;
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
			printf("(%d,%d): ",j,i);
			for (k = 0; k < zsize; k++) {
//				printf("%d:", k+ysize*j+xsize*i);
				printf("%d:", t);
				if (zsize == 1)
				printf("(%.2f,%.16f)\t", matrix[t][0], matrix[t][1]);
				else
				printf("(%f,%f)\t", matrix[t][0], matrix[t][1]);
				//printf("%f\n", matrix[t][0]);//, matrix[t][1]);
				t++;
			}
			printf("\n");
		}
		printf("\n");
	}
}




/*
 * Fills a host fftwf matrix with random numbers
 * --- USED FOR cpueM and cpurM ---
 */
void fillRandom(fftwf_complex ***matrix, int xsize, int ysize, int zsize) {
//	srand(time(0));
	int i, j, k;//,t=0;
//	printf("\n");
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
			for (k = 0; k < zsize; k++) {
//				printf("%d,%d,%d: \n",i,j,k);
				matrix[j][i][k][0] = 0.0f;//ifmodf(rand(), 10.0f) / 10.0f;
				matrix[j][i][k][1] = 0.0f;
//				matrix[j][i][k][0] = fmodf(rand(), 10.0f);
//				matrix[j][i][k][1] = fmodf(rand(), 10.0f);
			}
		}
	}
}

/*
 * Fills a host cufft matrix with random numbers
 * --- USED FOR fsrc ---
 */
void fillRandom(fftwf_complex *matrix, int size) {
//	srand(time(0));
	int i;
	for (i = 0; i < size; i++) {
		matrix[i][0] = 0.0f; //TODO remove fmodf(rand(), 10.0f) / 10.0f;
		matrix[i][1] = 0.0f; //TODO remove
//		matrix[i][0] = fmodf(rand(), 10.0f);
//		matrix[i][1] = fmodf(rand(), 10.0f);
	}
}

/*
 * Fills a host fftw matrix with random numbers
 * --- USED FOR src ---
 */
void fillRandom(fftw_complex *matrix, int size) {
	int i;
//	srand(time(0));
	for (i = 0; i < size; i++) {
		//TODO select appropriate range
		/*if(i < 512) {
		matrix[i][0] = 1.0; //TODO remove
		}
		else {
		matrix[i][0] = 0.0; //TODO remove
		}*/
		matrix[i][0] = 0.0; //TODO remove	
		matrix[i][1] = 0.0; //TODO remove
//		(*matrix)[i][0] = fmodf(rand(), 10.0f);
//		(*matrix)[i][1] = fmodf(rand(), 10.0f);
	}
}

/*
 * Fills a host fftw matrix with random numbers
 * --- USED FOR src ---
 */
void fillRandom(fftw_complex **matrix, int size) {
	int i;
	for (i = 0; i < size; i++) {
		matrix[0][i][0] = 0.0;	
		matrix[0][i][1] = 0.0;
	}
}

/*
 * Fills a host cufft matrix with random numbers
 * --- USED FOR elementMatrix ---
 */
void fillRandom(cufftComplex *matrix, int size) {
//	srand(time(0));
	int i;
	for (i = 0; i < size; i++) {
		matrix[i].x = 0.0f; //TODO remove
		matrix[i].y = 0.0f; //TODO remove
//		matrix[i].x = fmodf(rand(), 10.0f);
//		matrix[i].y = fmodf(rand(), 10.0f);
	}
}

/*
 * Fills a host cufft matrix with random numbers
 * --- USED FOR elementMatrix_h and correlationMatrix_h ---
 */
void fillRandom(cufftComplex **matrix, int xsize, int ysize) {
	int i,j;
	for (i = 0; i < ysize; i++) {
		for (j = 0; j < xsize; j++) {
			matrix[i][j].x = 1.0f;
			matrix[i][j].y = 0.0f;
		}
	}
}

/*
 * Fills the given matrix with random numbers.
 * (Used to create testdata for the correlation algorithm, not essential).
 */
void fillRandom2d(cufftComplex ***matrix, int xsize, int ysize, int paddedxsize, int paddedysize) {
	int i,j, xoffset, yoffset;
	srand(time(0));
	for (i = 0; i < paddedxsize; i++) {
		for (j = 0; j < paddedysize; j++) {
			matrix[0][i][j].x = 1.0f;
			matrix[0][i][j].y = 0.0f;
		}
	}
	xoffset = (paddedxsize-xsize)/2;
	yoffset = (paddedysize-ysize)/2;
	for (i = xoffset; i < xsize+xoffset; i++) {
		for (j = yoffset; j < ysize+yoffset; j++) {
			//TODO select appropriate range
			matrix[0][i][j].x = fmodf(rand(), 10.0f);
			matrix[0][i][j].y = fmodf(rand(), 10.0f);
		}
	}
}

/*
 * Returns the complex conjugate of 'z'
 * z = a + ib -> z* = a - ib.
 */
cufftComplex complexConjugate(cufftComplex z) {
	z.y = -z.y;
	return z;	
}

/*
 * Sums all the elements in a complex matrix.
 */
cufftComplex sum(cufftComplex **matrix, int xsize, int ysize) {
	int i, j;
	cufftComplex sum;
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			sum.x += matrix[i][j].x;
			sum.y += matrix[i][j].y;
		}
	}
	return sum;	
}

/*
 * Averages out the summed weight based on the size of the correlation grid used.
 */
cufftComplex average(cufftComplex sum, int xsize, int ysize) {
	sum.x /= (xsize * ysize);
	sum.y /= (xsize * ysize);
	return sum;	
}

//Round a / b to nearest higher integer value
int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

//Round a / b to nearest lower integer value
int iDivDown(int a, int b){
    return a / b;
}

//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
    return (a % b != 0) ?  (a - a % b + b) : a;
}

//Align a to nearest lower multiple of b
int iAlignDown(int a, int b){
    return a - a % b;
}


////////////////////////////////////////////////////////////////////////////////
// Device functions
////////////////////////////////////////////////////////////////////////////////
/*
 * Fills a matrix with given value
 */
__global__ void cuda_fill(cufftComplex *matrix, int size, float value) {
	//Multithread version
	int idx = threadIdx.x + blockIdx.x*blockDim.x + blockIdx.y*blockDim.x*gridDim.x;
	int i;
	for (i = idx; i < size; i+=blockDim.x*gridDim.x*gridDim.y) {
		matrix[i].x = value;
		matrix[i].y = value;
	}
/*	//Single thread version
	int i;
	for (i = 0; i < size; i++) {
//		matrix[idx+i].x = value;
//		matrix[idx+i].y = value;
		matrix[i].x = value;
		matrix[i].y = value;
	}
	*/
}

/*
 * Fills a matrix with given value
 */
__global__ void cuda_fill(fftw_complex *matrix, int size, double value) {
	int i;
	for (i = 0; i < size; i++) {
		matrix[i][0] = value;
		matrix[i][1] = value;
	}
}

/*
 * Returns the comlex conjugate of a cufftComplex
 */
__device__ cufftComplex cuda_complexConjugate(cufftComplex z) {
	z.y = -z.y;
	return z;	
}

/*
 * Sums the complex numbers in a matrix on the device
 */
__device__ cufftComplex cuda_sumMatrix(cufftComplex* matrix, int size) {
	int i;
	cufftComplex sum;
	for(i = 0; i < size; i++) {
		sum.x += matrix[i].x;
		sum.y += matrix[i].y;
	}
	return sum;
}

/*
 * Clears the two matrices given (fills them with zeroes).
 */
__global__ void cuda_clearMatrix(cufftComplex *matrix, size_t pitch) {
	cufftComplex *element = (cufftComplex*) ((char*)matrix + blockIdx.y*pitch) + blockIdx.x;
	element->x = 0.0f;
	element->y = 0.0f;
}

/*
 * Sets default startvalues for the correctionfactor matrix
 */
__global__ void cuda_setStartValues(cufftComplex *matrix, size_t pitch) {
	cufftComplex *element = (cufftComplex*) ((char*)matrix + blockIdx.y*pitch) + blockIdx.x;
	element->x = 1.0f; //Amplitude
	element->y = 0.0f; //Phase 
}

/*
 * Takes the square root of all elements in a matrix (used for eM_d and S)
 */
__global__ void cuda_sqrtMatrix(cufftComplex* matrix, int pitch) {
	cufftComplex *row = (cufftComplex*) ((char*)matrix + blockIdx.y*pitch) + blockIdx.x;
	row->x = sqrtf(row->x);
	row->y = sqrtf(row->x);
}

/*
 * Computes the difference between sM_d and sOM_d (original correctionfactors), 
 * saves result in rM_d.
 */
__global__ void cuda_computeError(	cufftComplex* sM_d, 
									cufftComplex* sOM_d,
									size_t sPitch, 
									size_t sOPitch,
									int xsize, 
									int ysize, 
									float *minError) {
	
	/*cufftComplex *sRow = 	(cufftComplex*) ((char*)sM_d 	+ 	blockIdx.y*sPitch) 	+ blockIdx.x;
	cufftComplex *sORow = 	(cufftComplex*) ((char*)sOM_d 	+ 	blockIdx.y*sOPitch) + blockIdx.x;
	cufftComplex *rRow = 	(cufftComplex*) ((char*)rM_d 	+ 	blockIdx.y*rPitch) 	+ blockIdx.x;
	rRow->x = sORow->x - sRow->x;
	rRow->y = sORow->y - sRow->y;*/
	
	float temp = 0;
	for (int j = 0; j < ysize; j++) {
		for (int k = 0; k < xsize; k++) {
			cufftComplex *sRow 	= (cufftComplex*) ((char*)sM_d 	+ 	j*sPitch) 	+ k;
			cufftComplex *sORow = (cufftComplex*) ((char*)sOM_d + 	j*sOPitch) 	+ k;
			temp += sqrtf(sORow->x*sORow->x + sORow->y*sORow->y) / sqrtf(sRow->x*sRow->x + sRow->y*sRow->y);
		}
	}
	temp /= ysize*xsize;
	temp = abs(temp - 1.0f);
	if (temp < minError[0]) {
		minError[0] = temp;
	}
	
	
	//TO USE:
	/*		// Calculate Error using GPU (slower than using CPU)
		cudaThreadSynchronize();
		cudaMemcpy2DAsync(	(cufftComplex*)((char*)sOM_d + s*sOPitch*YSIZE), sPitch, 
							(cufftComplex*)((char*)rM_d + s*sPitch*YSIZE), sPitch, 
							XSIZE*sizeof(cufftComplex), YSIZE, 
							cudaMemcpyDeviceToDevice, stream[s]);
		cudaThreadSynchronize();
		cuda_computeError<<<1,1>>>(	(cufftComplex*)((char*)sM_d + s*sPitch*YSIZE), 
									(cufftComplex*)((char*)sOM_d + s*sOPitch*YSIZE), 
									sPitch, 
									sOPitch, 
									xsize, 
									ysize, 
									minError_d);
		cudaThreadSynchronize();
		cudaMemcpy(	minError_h, minError_d, sizeof(float), cudaMemcpyDeviceToHost);
		cudaThreadSynchronize();

*/		
}

extern __shared__ cufftComplex sharedAvg[];
/*
 * Averages the values around the center freqency fc resulting from the fft 
 * of all values recorded by all elements of the transducer. Returns the average
 * value for each element to correlated over.
 * e.g. 64x64x1500 -> 64x64x1
 */
__global__ void cuda_averageFFTValues(cufftComplex* rM, cufftComplex* eM_d, const int xsize, const int ysize, const int zsize, size_t pitch, const int bandwidth, const int fc) {
//	int idx = blockIdx.x * blockDim.x + blockIdx.y * gridDim.x * blockDim.x + threadIdx.x;
		
	//Optimized from Harris implementation:
	// each thread loads one element from global to shared mem
	unsigned int tid = threadIdx.x;

	int idx = blockIdx.x * zsize + blockIdx.y * xsize * zsize + threadIdx.x*2;
	sharedAvg[tid].x = rM[idx+fc-IDIVUP(bandwidth, 2)].x + rM[idx+1+fc-IDIVUP(bandwidth, 2)].x;
	sharedAvg[tid].y = rM[idx+fc-IDIVUP(bandwidth, 2)].y + rM[idx+1+fc-IDIVUP(bandwidth, 2)].y;
	__syncthreads();
	// do reduction in shared mem
	if (blockDim.x % 32 == 0) { //Check if dividable by 32 or not all values will be summed.
		if (blockDim.x/2 > 32) { //Must be larger than 32 else for loop will not run and result will be wrong.
			for (unsigned int s=blockDim.x/2; s>32; s>>=1) { //s>>=1 divides s by 2^1
    			if (tid < s) {
        			sharedAvg[tid].x += sharedAvg[tid + s].x;
	        		sharedAvg[tid].y += sharedAvg[tid + s].y;
	    		}
    			__syncthreads();
			}
			if (tid < 32) {
				sharedAvg[tid].x += sharedAvg[tid + 32].x;
				sharedAvg[tid].y += sharedAvg[tid + 32].y;
				sharedAvg[tid].x += sharedAvg[tid + 16].x;
				sharedAvg[tid].y += sharedAvg[tid + 16].y;
				sharedAvg[tid].x += sharedAvg[tid + 8].x;
				sharedAvg[tid].y += sharedAvg[tid + 8].y;
				sharedAvg[tid].x += sharedAvg[tid + 4].x;
				sharedAvg[tid].y += sharedAvg[tid + 4].y;
				sharedAvg[tid].x += sharedAvg[tid + 2].x;
				sharedAvg[tid].y += sharedAvg[tid + 2].y;
				sharedAvg[tid].x += sharedAvg[tid + 1].x;
				sharedAvg[tid].y += sharedAvg[tid + 1].y;
			}
		}
		else { //Sum up all values under 32.
			for (unsigned int s=blockDim.x/2; s>0; s>>=1) {
				if (tid < s) {
					sharedAvg[tid].x += sharedAvg[tid + s].x;
    	    		sharedAvg[tid].y += sharedAvg[tid + s].y;
				}
				__syncthreads();
			}			
		}
	}
	else { //Not dividable by 32, will do as many as is possible until reaching an odd number
		int finalS = blockDim.x;
		for (unsigned int s=blockDim.x/2; s>32 && (s % 2 == 0); s>>=1) { //s>>=1 divides s by 2^1
			if (tid < s) {
    			sharedAvg[tid].x += sharedAvg[tid + s].x;
        		sharedAvg[tid].y += sharedAvg[tid + s].y;
    		}
			__syncthreads();
			finalS = s;
		}
		if (tid == 0) { //Will finish numbers under 32 or an odd number in serial fashion.
			for (unsigned int i = finalS-1; i > 0; i--) {
				sharedAvg[tid].x += sharedAvg[i].x;
        		sharedAvg[tid].y += sharedAvg[i].y;
			}	
		}
	}
/*
	unsigned int tid = threadIdx.x;
	sharedAvg[tid].x = rM[idx+FC-BANDWIDTH/2].x;
	sharedAvg[tid].y = rM[idx+FC-BANDWIDTH/2].y;*/
//	for (unsigned int s=1; s < blockDim.x; s *= 2) {
//		int index = 2 * s * tid;
//		if (index < blockDim.x) {
//			sharedAvg[index].x += sharedAvg[index + s].x;
//			sharedAvg[index].y += sharedAvg[index + s].y;
//		}
//		__syncthreads();
//	}
//	for (unsigned int s=1; s < blockDim.x; s *= 2) {
//		if (tid % (2*s) == 0) {
//			sharedAvg[tid].x += sharedAvg[tid + s].x;
//    	    sharedAvg[tid].y += sharedAvg[tid + s].y;
//		}
//		__syncthreads();
//	}
//	for(unsigned int s=1; s < blockDim.x; s *= 2) {
//	    if (tid % (2*s) == 0) {
//    	    sharedAvg[tid].x += sharedAvg[tid + s].x;
//    	    sharedAvg[tid].y += sharedAvg[tid + s].y;
//    	}
//    	__syncthreads();
//	}*/
	
	// write result for this block to global mem
	if (tid == 0) {
		sharedAvg[tid].x += rM[idx+fc+IDIVUP(bandwidth, 2)].x; //last value
		sharedAvg[tid].y += rM[idx+fc+IDIVUP(bandwidth, 2)].y; //last value
		cufftComplex *row = (cufftComplex*) ((char*)eM_d + blockIdx.y*pitch) + blockIdx.x;
		row->x = sharedAvg[tid].x/(bandwidth+1);
		row->y = sharedAvg[tid].y/(bandwidth+1);	
	}
	
	/*
	//Naive and slow implementation:
	int idx = blockIdx.x * zsize + blockIdx.y * xsize * zsize + threadIdx.x;
	sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH+1)].x = rM[idx+FC-BANDWIDTH/2].x;
	sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH+1)].y = rM[idx+FC-BANDWIDTH/2].y;
	if (threadIdx.x == 0) {
		int i;
		for (i = 1; i < BANDWIDTH + 1; i++) {
			sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1)].x += sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1) + i].x;
			sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1)].y += sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1) + i].y;
		}
		cufftComplex *row = (cufftComplex*) ((char*)eM_d + blockIdx.y*pitch) + blockIdx.x;
		row->x = sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1)].x/(BANDWIDTH + 1.0f);
		row->y = sharedAvg[threadIdx.x + blockIdx.y*(BANDWIDTH + 1)].y/(BANDWIDTH + 1.0f);
	}
	//Single thread implementation (not error checked):
	*//*
	int i,j,k;
	cufftComplex sum;
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			sum.x = 0.0f;
			sum.y = 0.0f;
			for (k = FC-BANDWIDTH/2; k < FC+BANDWIDTH/2; k++) { //should this be '<=' ?
				sum.x += rM[i + j*ysize + k*zsize].x;
				sum.y += rM[i + j*ysize + k*zsize].y;
			}
			sum.x /= BANDWIDTH;
			sum.y /= BANDWIDTH;
			cufftComplex *row = (cufftComplex*) ((char*)eM_d + j*pitch) + i;
			row->x = sum.x;
			row->y = sum.y;
		}
	}*/
	
}
