#ifndef GPUOPERATIONS_H_
#define GPUOPERATIONS_H_

#include <cufft.h>
#include <fftw3.h>

////////////////////////////////////////////////////////////////////////////////
// Kernel configuration
////////////////////////////////////////////////////////////////////////////////
//#ifndef FCBAND
//	#define FC 25 // Focus center freqency element
//	#define BANDWIDTH 14 // How many elements to average around FC, half on each side. (total of BANDWIDTH+1)
//	#define BANDWIDTHF 14.0f
//#endif
////////////////////////////////////////////////////////////////////////////////
// Integer operations
////////////////////////////////////////////////////////////////////////////////
//Round a / b to nearest higher integer value
#define IDIVUP(a, b) ( (a % b != 0) ? (a / b + 1) : (a / b) )
//Round a / b to nearest lower integer value
#define IDIVDOWN(a, b) ( a / b )
//Align a to nearest higher multiple of b
#define IALIGNUP(a, b) ( (a % b != 0) ?  (a - a % b + b) : a )
//Align a to nearest lower multiple of b
#define IALIGNDOWN(a, b) ( a - a % b )


////////////////////////////////////////////////////////////////////////////////
// Host functions
////////////////////////////////////////////////////////////////////////////////
void createMatrix(	cufftComplex **matrix, int xsize, int ysize, size_t &pitch);
void destroyMatrix(	cufftComplex **matrix);

void createMatrix(	cufftComplex ***matrix, int xsize, int ysize);
void destroyMatrix(	cufftComplex ***matrix, int ysize);

void printMatrix(	cufftComplex *matrix, int xsize, int ysize, int zsize);
void printMatrix(	cufftComplex *matrix, int xsize, int ysize);
void printMatrix(	cufftComplex *matrix, int size);
void printMatrix(	cufftComplex **matrix, int xsize, int ysize);
void printMatrix(	fftwf_complex ***matrix, int xsize, int ysize, int zsize);
void printMatrix(	fftw_complex *matrix, int xsize, int ysize, int zsize);


void fillRandom(	fftwf_complex ***matrix, int xsize, int ysize, int zsize);
void fillRandom(	fftwf_complex *matrix, int size);
void fillRandom(	fftw_complex *matrix, int size); //src
void fillRandom(	fftw_complex **matrix, int size); //src
void fillRandom(	cufftComplex *matrix, int size);
void fillRandom(	cufftComplex **matrix, int xsize, int ysize); //elementMatrix_h and correlationMatrix_h
void fillRandom2d(	cufftComplex ***matrix, int xsize, int ysize, int paddedxsize, int paddedysize);

cufftComplex complexConjugate(	cufftComplex complex);

cufftComplex sum(		cufftComplex **matrix, int xsize, int ysize);
cufftComplex average(	cufftComplex sum, int xsize, int ysize);

//Round a / b to nearest higher integer value
int iDivUp(int a, int b);

//Round a / b to nearest lower integer value
int iDivDown(int a, int b);

//Align a to nearest higher multiple of b
int iAlignUp(int a, int b);

//Align a to nearest lower multiple of b
int iAlignDown(int a, int b);

////////////////////////////////////////////////////////////////////////////////
// Device functions
////////////////////////////////////////////////////////////////////////////////
__global__ void cuda_fill(	cufftComplex *matrix, int size, float value	);

__global__ void cuda_fill(	fftw_complex *matrix, int size, double value);

__global__ void cuda_clearMatrix(	cufftComplex *matrix, size_t pitch);

__global__ void cuda_setStartValues(cufftComplex *matrix, size_t pitch);

__device__ cufftComplex cuda_complexConjugate(	cufftComplex z);

__device__ cufftComplex cuda_sumMatrix(	cufftComplex* matrix, 
										int size);

__global__ void cuda_sqrtMatrix(cufftComplex* matrix, int size);

__global__ void cuda_computeError(	cufftComplex* sM_d, 
									cufftComplex* sOM_d,
									size_t sPitch, 
									size_t sOPitch,
									int xsize, 
									int ysize, 
									float *minError);

__global__ void cuda_averageFFTValues(	cufftComplex* rM, 
										cufftComplex* eM_d, 
										const int xsize, 
										const int ysize, 
										const int zsize, 
										size_t pitch, 
										const int bandwidth, 
										const int fc);
										

#endif /*GPUOPERATIONS_H_*/
